#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import numpy as np
import pylab

import ltv_mpc

from ltv_mpc import solve_mpc


gravity = 9.81  # [m] / [s]^2


@dataclass
class HumanoidSteppingProblem:

    com_height: float = 0.8
    dsp_duration: float = 0.1
    end_pos: float = 0.3
    foot_length: float = 0.1
    horizon_duration: float = 2.5
    nb_timesteps: int = 16
    ssp_duration: float = 0.7
    start_pos: float = 0.0


def build_mpc_problem(problem: HumanoidSteppingProblem):
    T = problem.horizon_duration / problem.nb_timesteps
    nb_init_dsp_steps = int(round(problem.dsp_duration / T))
    nb_init_ssp_steps = int(round(problem.ssp_duration / T))
    nb_dsp_steps = int(round(problem.dsp_duration / T))
    state_matrix = np.array(
        [[1.0, T, T ** 2 / 2.0], [0.0, 1.0, T], [0.0, 0.0, 1.0]]
    )
    input_matrix = np.array([T ** 3 / 6.0, T ** 2 / 2.0, T])
    input_matrix = input_matrix.reshape((3, 1))
    zmp_from_state = np.array([1.0, 0.0, -problem.com_height / gravity])
    ineq_matrix = np.array([+zmp_from_state, -zmp_from_state])
    cur_max = problem.start_pos + 0.5 * problem.foot_length
    cur_min = problem.start_pos - 0.5 * problem.foot_length
    next_max = problem.end_pos + 0.5 * problem.foot_length
    next_min = problem.end_pos - 0.5 * problem.foot_length
    ineq_vector = [
        np.array([+1000.0, +1000.0])
        if i < nb_init_dsp_steps
        else np.array([+cur_max, -cur_min])
        if i - nb_init_dsp_steps <= nb_init_ssp_steps
        else np.array([+1000.0, +1000.0])
        if i - nb_init_dsp_steps - nb_init_ssp_steps < nb_dsp_steps
        else np.array([+next_max, -next_min])
        for i in range(problem.nb_timesteps)
    ]
    return ltv_mpc.Problem(
        transition_state_matrix=state_matrix,
        transition_input_matrix=input_matrix,
        ineq_state_matrix=ineq_matrix,
        ineq_input_matrix=None,
        ineq_vector=ineq_vector,
        initial_state=np.array([problem.start_pos, 0.0, 0.0]),
        goal_state=np.array([problem.end_pos, 0.0, 0.0]),
        nb_timesteps=problem.nb_timesteps,
        terminal_cost_weight=1.0,
        stage_state_cost_weight=None,
        stage_input_cost_weight=1e-3,
    )


def plot_mpc_solution(stepping_problem, mpc, solution):
    t = np.linspace(
        0.0,
        stepping_problem.horizon_duration,
        stepping_problem.nb_timesteps + 1,
    )
    X = solution.stacked_states
    zmp_from_state = np.array(
        [1.0, 0.0, -stepping_problem.com_height / gravity]
    )
    zmp = X.dot(zmp_from_state)
    pos = X[:, 0]
    zmp_min = [x[0] if abs(x[0]) < 10 else None for x in mpc.ineq_vector]
    zmp_max = [-x[1] if abs(x[1]) < 10 else None for x in mpc.ineq_vector]
    zmp_min.append(zmp_min[-1])
    zmp_max.append(zmp_max[-1])
    pylab.ion()
    pylab.clf()
    pylab.plot(t, pos)
    pylab.plot(t, zmp, "r-")
    pylab.plot(t, zmp_min, "g:")
    pylab.plot(t, zmp_max, "b:")
    pylab.grid(True)


if __name__ == "__main__":
    stepping_problem = HumanoidSteppingProblem()
    mpc_problem = build_mpc_problem(stepping_problem)
    solution = solve_mpc(mpc_problem)
    plot_mpc_solution(stepping_problem, mpc_problem, solution)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Linear time-variant model predictive control in Python.
"""

from .problem import Problem
from .solution import Solution
from .solve_mpc import solve_mpc

__all__ = [
    "Problem",
    "Solution",
    "solve_mpc",
]

__version__ = "0.7.0"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from dataclasses import dataclass

import numpy as np

import ltv_mpc

from ltv_mpc.solve_mpc import build_qp
from ltv_mpc import solve_mpc


gravity = 9.81  # [m] / [s]^2


@dataclass
class HumanoidSteppingProblem:

    com_height: float = 0.8
    dsp_duration: float = 0.1
    end_pos: float = 0.3
    foot_length: float = 0.1
    horizon_duration: float = 2.5
    nb_timesteps: int = 16
    ssp_duration: float = 0.7
    start_pos: float = 0.0


class TestHumanoid(unittest.TestCase):
    def setUp(self):
        """
        Prepare test fixture.
        """
        problem = HumanoidSteppingProblem()
        T = problem.horizon_duration / problem.nb_timesteps
        nb_init_dsp_steps = int(round(problem.dsp_duration / T))
        nb_init_ssp_steps = int(round(problem.ssp_duration / T))
        nb_dsp_steps = int(round(problem.dsp_duration / T))
        state_matrix = np.array(
            [[1.0, T, T ** 2 / 2.0], [0.0, 1.0, T], [0.0, 0.0, 1.0]]
        )
        input_matrix = np.array([T ** 3 / 6.0, T ** 2 / 2.0, T])
        input_matrix = input_matrix.reshape((3, 1))
        zmp_from_state = np.array([1.0, 0.0, -problem.com_height / gravity])
        ineq_matrix = np.array([+zmp_from_state, -zmp_from_state])
        cur_max = problem.start_pos + 0.5 * problem.foot_length
        cur_min = problem.start_pos - 0.5 * problem.foot_length
        next_max = problem.end_pos + 0.5 * problem.foot_length
        next_min = problem.end_pos - 0.5 * problem.foot_length
        ineq_vector = [
            np.array([+1000.0, +1000.0])
            if i < nb_init_dsp_steps
            else np.array([+cur_max, -cur_min])
            if i - nb_init_dsp_steps <= nb_init_ssp_steps
            else np.array([+1000.0, +1000.0])
            if i - nb_init_dsp_steps - nb_init_ssp_steps < nb_dsp_steps
            else np.array([+next_max, -next_min])
            for i in range(problem.nb_timesteps)
        ]
        mpc_problem = ltv_mpc.Problem(
            transition_state_matrix=state_matrix,
            transition_input_matrix=input_matrix,
            ineq_state_matrix=ineq_matrix,
            ineq_input_matrix=None,
            ineq_vector=ineq_vector,
            initial_state=np.array([problem.start_pos, 0.0, 0.0]),
            goal_state=np.array([problem.end_pos, 0.0, 0.0]),
            nb_timesteps=problem.nb_timesteps,
            terminal_cost_weight=1.0,
            stage_state_cost_weight=None,
            stage_input_cost_weight=1e-3,
        )
        #
        self.stepping_problem = problem
        self.problem = mpc_problem

    def test_build_qp(self):
        qp = build_qp(self.problem)
        G = qp.ineq_matrix
        self.assertNotAlmostEqual(np.linalg.norm(G[0]), 0.0)
        self.assertNotAlmostEqual(np.linalg.norm(G[1]), 0.0)

    def test_solve_mpc(self):
        solution = solve_mpc(self.problem)
        N = self.problem.nb_timesteps
        input_dim = self.problem.input_dim
        state_dim = self.problem.state_dim
        self.assertEqual(
            solution.stacked_inputs.flatten().shape, (N * input_dim,)
        )
        self.assertEqual(
            solution.stacked_states.flatten().shape, ((N + 1) * state_dim,)
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .problem import Problem


class Solution:

    """
    State and input trajectories that optimize a given
    :class:`ltv_mpc.problem.Problem`.

    Attributes:
        stacked_inputs: Stacked vector of inputs :math:`u_k` for
            :math:`k \\in \\{0, \\ldots, N - 1\\}`.
    """

    stacked_inputs: np.ndarray

    def __init__(self, problem: Problem, stacked_inputs: np.ndarray):
        """
        Test.
        """
        self.problem = problem
        self.stacked_inputs = stacked_inputs
        self.__stacked_states = None

    @property
    def stacked_states(self):
        """
        Stacked vector of states :math:`x_k` for
        :math:`k \\in \\{0, \\ldots, N\\}`, with :math:`N` the number of
        timesteps.

        Note:
            The time complexity of calling this property is :math:`O(N)` the
            first time, then :math:`O(1)` as the result is memoized.
        """
        if self.__stacked_states is not None:
            return self.__stacked_states
        X = np.zeros((self.problem.nb_timesteps + 1, self.problem.state_dim))
        X[0] = self.problem.initial_state
        for k in range(self.problem.nb_timesteps):
            A = self.problem.get_transition_state_matrix(k)
            B = self.problem.get_transition_input_matrix(k)
            X[k + 1] = A.dot(X[k]) + B.dot(self.stacked_inputs[k])
        self.__stacked_states = X
        return X



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Union

import numpy as np

from scipy.sparse import csc_matrix
from qpsolvers import solve_qp

from .problem import Problem
from .solution import Solution


@dataclass
class QuadraticProgram:

    """
    Quadratic program (QP) with inequality constraints.
    """

    cost_matrix: Union[np.ndarray, csc_matrix]
    cost_vector: np.ndarray
    ineq_matrix: Union[np.ndarray, csc_matrix]
    ineq_vector: np.ndarray


def build_qp(problem: Problem, sparse: bool = False) -> QuadraticProgram:
    """
    Build the quadratic program corresponding to an LTV-MPC problem.

    Args:
        problem: Model predictive control problem.
        sparse: Whether to use sparse or dense matrices in the output quadratic
            program. Enable it if you are calling a sparse solver afterwards.

    Returns:
        Quadratic program representing the input problem.

    Notes:
        In numerical analysis, there are three classes of methods to solve
        boundary value problems: single shooting, multiple shooting and
        collocation. The QP built by this function implements a `single
        shooting method <https://en.wikipedia.org/wiki/Shooting_method>`_.
    """
    input_dim = problem.input_dim
    state_dim = problem.state_dim
    stacked_input_dim = problem.input_dim * problem.nb_timesteps

    phi = np.eye(state_dim)
    psi = np.zeros((state_dim, stacked_input_dim))
    G_list, h_list = [], []
    phi_list, psi_list = [], []
    for k in range(problem.nb_timesteps):
        # Loop invariant: x == psi * U + phi * x_init
        if problem.stage_state_cost_weight is not None:
            phi_list.append(phi)
            psi_list.append(psi)
        A = problem.get_transition_state_matrix(k)
        B = problem.get_transition_input_matrix(k)
        C = problem.get_ineq_state_matrix(k)
        D = problem.get_ineq_input_matrix(k)
        e = problem.get_ineq_vector(k)
        G = np.zeros((e.shape[0], stacked_input_dim))
        h = e if C is None else e - np.dot(C.dot(phi), problem.initial_state)
        input_slice = slice(k * input_dim, (k + 1) * input_dim)
        if D is not None:
            # we rely on G == 0 to avoid a slower +=
            G[:, input_slice] = D
        if C is not None:
            G += C.dot(psi)
        if k == 0 and D is None:  # corner case, input has no effect
            assert np.all(h >= 0.0)
        else:  # regular case, G is non-zero
            G_list.append(G)
            h_list.append(h)
        phi = A.dot(phi)
        psi = A.dot(psi)
        psi[:, input_slice] = B

    P = problem.stage_input_cost_weight * np.eye(stacked_input_dim)
    q = np.zeros(stacked_input_dim)
    if (
        problem.terminal_cost_weight is not None
        and problem.terminal_cost_weight > 1e-10
    ):
        c = np.dot(phi, problem.initial_state) - problem.goal_state
        P += problem.terminal_cost_weight * np.dot(psi.T, psi)
        q += problem.terminal_cost_weight * np.dot(c.T, psi)
    if (
        problem.stage_state_cost_weight is not None
        and problem.stage_state_cost_weight > 1e-10
    ):
        Phi = np.vstack(phi_list)
        Psi = np.vstack(psi_list)
        X_goal = np.hstack([problem.goal_state] * problem.nb_timesteps)
        c = np.dot(Phi, problem.initial_state) - X_goal
        P += problem.stage_state_cost_weight * np.dot(Psi.T, Psi)
        q += problem.stage_state_cost_weight * np.dot(c.T, Psi)

    G = np.vstack(G_list)
    h = np.hstack(h_list)
    if sparse:
        P = csc_matrix(P)
        G = csc_matrix(G)
    return QuadraticProgram(
        cost_matrix=P, cost_vector=q, ineq_matrix=G, ineq_vector=h
    )


def solve_mpc(problem: Problem, sparse: bool = False, **kwargs) -> Solution:
    """
    Solve a linear time-invariant model predictive control problem.

    Args:
        problem: Model predictive control problem to solve.
        sparse: Whether to use sparse or dense matrices in the output quadratic
            program. Enable it if the QP solver is sparse (e.g. OSQP).

    Returns:
        Solution to the problem, if found.

    Note:
        Keyword arguments are passed to the QP solver via the `solve_qp`_
        function. In particular, the ``solver`` string can be set to select a
        different QP solver.

    .. _solve_qp:
        https://scaron.info/doc/qpsolvers/quadratic-programming.html#qpsolvers.solve_qp
    """
    qp = build_qp(problem, sparse=sparse)
    U = solve_qp(
        qp.cost_matrix,
        qp.cost_vector,
        qp.ineq_matrix,
        qp.ineq_vector,
        **kwargs
    )
    U = U.reshape((problem.nb_timesteps, problem.input_dim))
    return Solution(problem, U)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pylab

from ltv_mpc import Problem
from ltv_mpc import solve_mpc


if __name__ == "__main__":
    # Double integrator LTV system
    horizon_duration = 1.0
    nb_timesteps = 16
    T = horizon_duration / nb_timesteps
    A = np.array([[1.0, T, T ** 2 / 2.0], [0.0, 1.0, T], [0.0, 0.0, 1.0]])
    B = np.array([T ** 3 / 6.0, T ** 2 / 2.0, T]).reshape((3, 1))

    # Acceleration limits
    accel_from_state = np.array([0.0, 0.0, 1.0])
    max_accel = 3.0  # [m] / [s] / [s]
    ineq_matrix = np.vstack([+accel_from_state, -accel_from_state])
    ineq_vector = np.array([+max_accel, +max_accel])

    # Complete LTV problem
    initial_pos = 0.0
    goal_pos = 1.0
    problem = Problem(
        transition_state_matrix=A,
        transition_input_matrix=B,
        ineq_state_matrix=ineq_matrix,
        ineq_input_matrix=None,
        ineq_vector=ineq_vector,
        initial_state=np.array([initial_pos, 0.0, 0.0]),
        goal_state=np.array([goal_pos, 0.0, 0.0]),
        nb_timesteps=nb_timesteps,
        terminal_cost_weight=1.0,
        stage_state_cost_weight=None,
        stage_input_cost_weight=1e-6,
    )

    # Solve our LTV problem
    solution = solve_mpc(problem, solver="osqp")

    # Plot solution
    t = np.linspace(0.0, horizon_duration, nb_timesteps + 1)
    X = solution.stacked_states
    positions, velocities, accelerations = X[:, 0], X[:, 1], X[:, 2]
    pylab.ion()
    pylab.clf()
    pylab.plot(t, positions)
    pylab.plot(t, velocities)
    pylab.plot(t, accelerations)
    pylab.grid(True)
    pylab.legend(("position", "velocity", "acceleration"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

import numpy as np


class Problem:

    """
    Linear time-variant model predictive control problem.

    The discretized dynamics of a linear system are described by:

    .. math::

        x_{k+1} = A x_k + B u_k

    where :math:`x` is assumed to be the first-order state of a configuration
    variable :math:`p`, i.e., it stacks both the position :math:`p` and its
    time-derivative :math:`\\dot{p}`. Meanwhile, the system is linearly
    constrained by:

    .. math::

        x_0 & = x_\\mathrm{init} \\\\
        \\forall k, \\ C_k x_k + D_k u_k & \\leq e_k \\\\

    The output control law minimizes a weighted combination of two types of
    costs:

    - Terminal state error
        :math:`\\|x_\\mathrm{nb\\_steps} - x_\\mathrm{goal}\\|^2`
        with weight :math:`w_{xt}`.
    - Cumulated state error:
        :math:`\\sum_k \\|x_k - x_\\mathrm{goal}\\|^2`
        with weight :math:`w_{xc}`.
    - Cumulated control costs:
        :math:`\\sum_k \\|u_k\\|^2`
        with weight :math:`w_{u}`.

    Attributes:
        transition_state_matrix: State linear dynamics matrix.
        transition_input_matrix: Control linear dynamics matrix.
        ineq_state_matrix : Constraint matrix on state variables. When this
            argument is an array, the same matrix `C` is applied at each step
            `k`. When it is ``None``, the null matrix is applied.
        ineq_input_matrix : Constraint matrix on control variables. When this
            argument is an array, the same matrix `D` is applied at each step
            `k`. When it is ``None``, the null matrix is applied.
        ineq_vector : Constraint vector. When this argument is an array, the
            same vector `e` is applied at each step `k`.
        initial_state: Initial state as stacked position and velocity.
        goal_state: Goal state as stacked position and velocity.
        nb_timesteps: Number of discretization steps in the preview window.
        terminal_cost_weight: Weight on terminal state cost, or ``None`` to
            disable.
        stage_state_cost_weight: Weight on cumulated state costs, or ``None``
            to disable (default).
        stage_input_cost_weight: Weight on cumulated control costs.
        input_dim: Dimension of an input vector.
        state_dim: Dimension of a state vector.
    """

    transition_state_matrix: Union[np.ndarray, List[np.ndarray]]
    transition_input_matrix: Union[np.ndarray, List[np.ndarray]]
    ineq_state_matrix: Union[None, np.ndarray, List[np.ndarray]]
    ineq_input_matrix: Union[None, np.ndarray, List[np.ndarray]]
    ineq_vector: Union[np.ndarray, List[np.ndarray]]
    initial_state: np.ndarray
    goal_state: np.ndarray
    nb_timesteps: int
    terminal_cost_weight: Optional[float]
    stage_state_cost_weight: Optional[float]
    stage_input_cost_weight: float

    input_dim: int
    state_dim: int

    def __init__(
        self,
        transition_state_matrix: Union[np.ndarray, List[np.ndarray]],
        transition_input_matrix: Union[np.ndarray, List[np.ndarray]],
        ineq_state_matrix: Union[None, np.ndarray, List[np.ndarray]],
        ineq_input_matrix: Union[None, np.ndarray, List[np.ndarray]],
        ineq_vector: Union[np.ndarray, List[np.ndarray]],
        initial_state: np.ndarray,
        goal_state: np.ndarray,
        nb_timesteps: int,
        terminal_cost_weight: Optional[float],
        stage_state_cost_weight: Optional[float],
        stage_input_cost_weight: float,
    ) -> None:
        assert (
            ineq_state_matrix is not None or ineq_input_matrix is not None
        ), "use LQR for unconstrained case"
        assert (
            stage_input_cost_weight > 0.0
        ), "non-negative control weight needed for regularization"
        assert (
            terminal_cost_weight is not None
            or stage_state_cost_weight is not None
        ), "set either wxt or wxc"
        self.transition_state_matrix = transition_state_matrix
        self.transition_input_matrix = transition_input_matrix
        self.ineq_state_matrix = ineq_state_matrix
        self.ineq_input_matrix = ineq_input_matrix
        self.ineq_vector = ineq_vector
        self.initial_state = initial_state
        self.goal_state = goal_state
        self.nb_timesteps = nb_timesteps
        self.terminal_cost_weight = terminal_cost_weight
        self.stage_state_cost_weight = stage_state_cost_weight
        self.stage_input_cost_weight = stage_input_cost_weight
        if isinstance(transition_input_matrix, np.ndarray):
            self.input_dim = transition_input_matrix.shape[1]
        else:  # isinstance(transition_input_matrix, List[np.ndarray])
            self.input_dim = transition_input_matrix[0].shape[1]
        if isinstance(transition_state_matrix, np.ndarray):
            self.state_dim = transition_state_matrix.shape[1]
        else:  # isinstance(transition_state_matrix, List[np.ndarray])
            self.state_dim = transition_state_matrix[0].shape[1]

    def get_transition_state_matrix(self, k):
        return (
            self.transition_state_matrix[k]
            if isinstance(self.transition_state_matrix, list)
            else self.transition_state_matrix
        )

    def get_transition_input_matrix(self, k):
        return (
            self.transition_input_matrix[k]
            if isinstance(self.transition_input_matrix, list)
            else self.transition_input_matrix
        )

    def get_ineq_state_matrix(self, k):
        return (
            self.ineq_state_matrix[k]
            if isinstance(self.ineq_state_matrix, list)
            else self.ineq_state_matrix
        )

    def get_ineq_input_matrix(self, k):
        return (
            self.ineq_input_matrix[k]
            if isinstance(self.ineq_input_matrix, list)
            else self.ineq_input_matrix
        )

    def get_ineq_vector(self, k):
        return (
            self.ineq_vector[k]
            if isinstance(self.ineq_vector, list)
            else self.ineq_vector
        )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ltv_mpc']

package_data = \
{'': ['*']}

install_requires = \
['qpsolvers >=1.8.0']

extras_require = \
{'doc': ['sphinx']}

setup(name='ltv-mpc',
      version='0.7.0',
      description='Linear time-variant model predictive control in Python.',
      author=None,
      author_email='Stéphane Caron <stephane.caron@normalesup.org>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )
